@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- encoding: utf-8 -*-
 
-# Copyright (c) 2002-2020 "Neo4j,"
+# Copyright (c) "Neo4j"
 # Neo4j Sweden AB [http://neo4j.com]
 #
 # This file is part of Neo4j.
@@ -18,47 +18,43 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from collections import deque
+from enum import Enum
+from logging import getLogger
 from ssl import SSLSocket
-from time import perf_counter
-from neo4j.api import (
-    Version,
-    READ_ACCESS,
-    DEFAULT_DATABASE,
-    SYSTEM_DATABASE,
-)
-from neo4j.io._common import (
-    Inbox,
-    Outbox,
-    Response,
-    InitResponse,
-    CommitResponse,
-)
-from neo4j.meta import get_user_agent
-from neo4j.exceptions import (
-    AuthError,
-    ServiceUnavailable,
-    DatabaseUnavailable,
-    NotALeader,
-    ForbiddenOnReadOnlyDatabase,
-    SessionExpired,
-)
+
 from neo4j._exceptions import (
-    BoltIncompleteCommitError,
+    BoltError,
     BoltProtocolError,
 )
-from neo4j.packstream import (
-    Unpacker,
-    Packer,
+from neo4j.api import (
+    READ_ACCESS,
+    SYSTEM_DATABASE,
+    Version,
+)
+from neo4j.exceptions import (
+    ConfigurationError,
+    DatabaseUnavailable,
+    DriverError,
+    ForbiddenOnReadOnlyDatabase,
+    Neo4jError,
+    NotALeader,
+    ServiceUnavailable,
 )
 from neo4j.io import (
     Bolt,
-    BoltPool,
+    check_supported_server_product,
+)
+from neo4j.io._common import (
+    CommitResponse,
+    InitResponse,
+    Response,
+)
+from neo4j.io._bolt3 import (
+    ServerStateManager,
+    ServerStates,
 )
-from neo4j.api import ServerInfo
-from neo4j.addressing import Address
 
-from logging import getLogger
+
 log = getLogger("neo4j")
 
 
@@ -70,60 +66,29 @@ class Bolt4x0(Bolt):
 
     PROTOCOL_VERSION = Version(4, 0)
 
-    # The socket
-    in_use = False
+    supports_multiple_results = True
 
-    # The socket
-    _closed = False
+    supports_multiple_databases = True
 
-    # The socket
-    _defunct = False
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self._server_state_manager = ServerStateManager(
+            ServerStates.CONNECTED, on_change=self._on_server_state_change
+        )
 
-    #: The pool of which this connection is a member
-    pool = None
+    def _on_server_state_change(self, old_state, new_state):
+        log.debug("[#%04X]  State: %s > %s", self.local_port,
+                  old_state.name, new_state.name)
 
-    def __init__(self, unresolved_address, sock, max_connection_lifetime, *, auth=None, user_agent=None, routing_context=None):
-        self.unresolved_address = unresolved_address
-        self.socket = sock
-        self.server_info = ServerInfo(Address(sock.getpeername()), self.PROTOCOL_VERSION)
-        self.outbox = Outbox()
-        self.inbox = Inbox(self.socket, on_error=self._set_defunct)
-        self.packer = Packer(self.outbox)
-        self.unpacker = Unpacker(self.inbox)
-        self.responses = deque()
-        self._max_connection_lifetime = max_connection_lifetime  # self.pool_config.max_connection_lifetime
-        self._creation_timestamp = perf_counter()
-        self.supports_multiple_results = True
-        self.supports_multiple_databases = True
-        self._is_reset = True
-        self.routing_context = routing_context
-
-        # Determine the user agent
-        if user_agent:
-            self.user_agent = user_agent
-        else:
-            self.user_agent = get_user_agent()
-
-        # Determine auth details
-        if not auth:
-            self.auth_dict = {}
-        elif isinstance(auth, tuple) and 2 <= len(auth) <= 3:
-            from neo4j import Auth
-            self.auth_dict = vars(Auth("basic", *auth))
-        else:
-            try:
-                self.auth_dict = vars(auth)
-            except (KeyError, TypeError):
-                raise AuthError("Cannot determine auth details from %r" % auth)
-
-        # Check for missing password
-        try:
-            credentials = self.auth_dict["credentials"]
-        except KeyError:
-            pass
-        else:
-            if credentials is None:
-                raise AuthError("Password cannot be None")
+    @property
+    def is_reset(self):
+        # We can't be sure of the server's state if there are still pending
+        # responses. Unless the last message we sent was RESET. In that case
+        # the server state will always be READY when we're done.
+        if (self.responses and self.responses[-1]
+                and self.responses[-1].message == "reset"):
+            return True
+        return self._server_state_manager.state == ServerStates.READY
 
     @property
     def encrypted(self):
@@ -133,13 +98,6 @@ class Bolt4x0(Bolt):
     def der_encoded_server_certificate(self):
         return self.socket.getpeercert(binary_form=True)
 
-    @property
-    def local_port(self):
-        try:
-            return self.socket.getsockname()[1]
-        except IOError:
-            return 0
-
     def get_base_headers(self):
         return {
             "user_agent": self.user_agent,
@@ -153,11 +111,56 @@ class Bolt4x0(Bolt):
             logged_headers["credentials"] = "*******"
         log.debug("[#%04X]  C: HELLO %r", self.local_port, logged_headers)
         self._append(b"\x01", (headers,),
-                     response=InitResponse(self, on_success=self.server_info._update_metadata))
+                     response=InitResponse(self, "hello",
+                     on_success=self.server_info.update))
         self.send_all()
         self.fetch_all()
+        check_supported_server_product(self.server_info.agent)
 
-    def run(self, query, parameters=None, mode=None, bookmarks=None, metadata=None, timeout=None, db=None, **handlers):
+    def route(self, database=None, imp_user=None, bookmarks=None):
+        if imp_user is not None:
+            raise ConfigurationError(
+                "Impersonation is not supported in Bolt Protocol {!r}. "
+                "Trying to impersonate {!r}.".format(
+                    self.PROTOCOL_VERSION, imp_user
+                )
+            )
+        metadata = {}
+        records = []
+
+        if database is None:  # default database
+            self.run(
+                "CALL dbms.routing.getRoutingTable($context)",
+                {"context": self.routing_context},
+                mode="r",
+                bookmarks=bookmarks,
+                db=SYSTEM_DATABASE,
+                on_success=metadata.update
+            )
+        else:
+            self.run(
+                "CALL dbms.routing.getRoutingTable($context, $database)",
+                {"context": self.routing_context, "database": database},
+                mode="r",
+                bookmarks=bookmarks,
+                db=SYSTEM_DATABASE,
+                on_success=metadata.update
+            )
+        self.pull(on_success=metadata.update, on_records=records.extend)
+        self.send_all()
+        self.fetch_all()
+        routing_info = [dict(zip(metadata.get("fields", ()), values)) for values in records]
+        return routing_info
+
+    def run(self, query, parameters=None, mode=None, bookmarks=None,
+            metadata=None, timeout=None, db=None, imp_user=None, **handlers):
+        if imp_user is not None:
+            raise ConfigurationError(
+                "Impersonation is not supported in Bolt Protocol {!r}. "
+                "Trying to impersonate {!r}.".format(
+                    self.PROTOCOL_VERSION, imp_user
+                )
+            )
         if not parameters:
             parameters = {}
         extra = {}
@@ -183,48 +186,34 @@ class Bolt4x0(Bolt):
         fields = (query, parameters, extra)
         log.debug("[#%04X]  C: RUN %s", self.local_port, " ".join(map(repr, fields)))
         if query.upper() == u"COMMIT":
-            self._append(b"\x10", fields, CommitResponse(self, **handlers))
+            self._append(b"\x10", fields, CommitResponse(self, "run",
+                                                         **handlers))
         else:
-            self._append(b"\x10", fields, Response(self, **handlers))
-        self._is_reset = False
-
-    def run_get_routing_table(self, on_success, on_failure, database=DEFAULT_DATABASE):
-        if database == DEFAULT_DATABASE:
-            self.run(
-                "CALL dbms.routing.getRoutingTable($context)",
-                {"context": self.routing_context},
-                mode="r",
-                db=SYSTEM_DATABASE,
-                on_success=on_success,
-                on_failure=on_failure,
-            )
-        else:
-            self.run(
-                "CALL dbms.routing.getRoutingTable($context, $database)",
-                {"context": self.routing_context, "database": database},
-                mode="r",
-                db=SYSTEM_DATABASE,
-                on_success=on_success,
-                on_failure=on_failure,
-            )
+            self._append(b"\x10", fields, Response(self, "run", **handlers))
 
     def discard(self, n=-1, qid=-1, **handlers):
         extra = {"n": n}
         if qid != -1:
             extra["qid"] = qid
         log.debug("[#%04X]  C: DISCARD %r", self.local_port, extra)
-        self._append(b"\x2F", (extra,), Response(self, **handlers))
+        self._append(b"\x2F", (extra,), Response(self, "discard", **handlers))
 
     def pull(self, n=-1, qid=-1, **handlers):
         extra = {"n": n}
         if qid != -1:
             extra["qid"] = qid
         log.debug("[#%04X]  C: PULL %r", self.local_port, extra)
-        self._append(b"\x3F", (extra,), Response(self, **handlers))
-        self._is_reset = False
+        self._append(b"\x3F", (extra,), Response(self, "pull", **handlers))
 
     def begin(self, mode=None, bookmarks=None, metadata=None, timeout=None,
-              db=None, **handlers):
+              db=None, imp_user=None, **handlers):
+        if imp_user is not None:
+            raise ConfigurationError(
+                "Impersonation is not supported in Bolt Protocol {!r}. "
+                "Trying to impersonate {!r}.".format(
+                    self.PROTOCOL_VERSION, imp_user
+                )
+            )
         extra = {}
         if mode in (READ_ACCESS, "r"):
             extra["mode"] = "r"  # It will default to mode "w" if nothing is specified
@@ -246,28 +235,15 @@ class Bolt4x0(Bolt):
             except TypeError:
                 raise TypeError("Timeout must be specified as a number of seconds")
         log.debug("[#%04X]  C: BEGIN %r", self.local_port, extra)
-        self._append(b"\x11", (extra,), Response(self, **handlers))
-        self._is_reset = False
+        self._append(b"\x11", (extra,), Response(self, "begin", **handlers))
 
     def commit(self, **handlers):
         log.debug("[#%04X]  C: COMMIT", self.local_port)
-        self._append(b"\x12", (), CommitResponse(self, **handlers))
+        self._append(b"\x12", (), CommitResponse(self, "commit", **handlers))
 
     def rollback(self, **handlers):
         log.debug("[#%04X]  C: ROLLBACK", self.local_port)
-        self._append(b"\x13", (), Response(self, **handlers))
-
-    def _append(self, signature, fields=(), response=None):
-        """ Add a message to the outgoing queue.
-
-        :arg signature: the signature of the message
-        :arg fields: the fields of the message as a tuple
-        :arg response: a response object to handle callbacks
-        """
-        self.packer.pack_struct(signature, fields)
-        self.outbox.chunk()
-        self.outbox.chunk()
-        self.responses.append(response)
+        self._append(b"\x13", (), Response(self, "rollback", **handlers))
 
     def reset(self):
         """ Add a RESET message to the outgoing queue, send
@@ -278,42 +254,12 @@ class Bolt4x0(Bolt):
             raise BoltProtocolError("RESET failed %r" % metadata, self.unresolved_address)
 
         log.debug("[#%04X]  C: RESET", self.local_port)
-        self._append(b"\x0F", response=Response(self, on_failure=fail))
+        self._append(b"\x0F", response=Response(self, "reset", on_failure=fail))
         self.send_all()
         self.fetch_all()
-        self._is_reset = True
-
-    def _send_all(self):
-        data = self.outbox.view()
-        if data:
-            self.socket.sendall(data)
-            self.outbox.clear()
-
-    def send_all(self):
-        """ Send all queued messages to the server.
-        """
-        if self.closed():
-            raise ServiceUnavailable("Failed to write to closed connection {!r} ({!r})".format(
-                self.unresolved_address, self.server_info.address))
-
-        if self.defunct():
-            raise ServiceUnavailable("Failed to write to defunct connection {!r} ({!r})".format(
-                self.unresolved_address, self.server_info.address))
-
-        try:
-            self._send_all()
-        except (IOError, OSError) as error:
-            log.error("Failed to write data to connection "
-                      "{!r} ({!r}); ({!r})".
-                      format(self.unresolved_address,
-                             self.server_info.address,
-                             "; ".join(map(repr, error.args))))
-            if self.pool:
-                self.pool.deactivate(address=self.unresolved_address)
-            raise
 
     def fetch_message(self):
-        """ Receive at least one message from the server, if available.
+        """ Receive at most one message from the server, if available.
 
         :return: 2-tuple of number of detail messages and number of summary
                  messages fetched
@@ -330,17 +276,7 @@ class Bolt4x0(Bolt):
             return 0, 0
 
         # Receive exactly one message
-        try:
-            details, summary_signature, summary_metadata = next(self.inbox)
-        except (IOError, OSError) as error:
-            log.error("Failed to read data from connection "
-                      "{!r} ({!r}); ({!r})".
-                      format(self.unresolved_address,
-                             self.server_info.address,
-                             "; ".join(map(repr, error.args))))
-            if self.pool:
-                self.pool.deactivate(address=self.unresolved_address)
-            raise
+        details, summary_signature, summary_metadata = next(self.inbox)
 
         if details:
             log.debug("[#%04X]  S: RECORD * %d", self.local_port, len(details))  # Do not log any data
@@ -353,12 +289,15 @@ class Bolt4x0(Bolt):
         response.complete = True
         if summary_signature == b"\x70":
             log.debug("[#%04X]  S: SUCCESS %r", self.local_port, summary_metadata)
+            self._server_state_manager.transition(response.message,
+                                                  summary_metadata)
             response.on_success(summary_metadata or {})
         elif summary_signature == b"\x7E":
             log.debug("[#%04X]  S: IGNORED", self.local_port)
             response.on_ignored(summary_metadata or {})
         elif summary_signature == b"\x7F":
             log.debug("[#%04X]  S: FAILURE %r", self.local_port, summary_metadata)
+            self._server_state_manager.state = ServerStates.FAILED
             try:
                 response.on_failure(summary_metadata or {})
             except (ServiceUnavailable, DatabaseUnavailable):
@@ -369,68 +308,16 @@ class Bolt4x0(Bolt):
                 if self.pool:
                     self.pool.on_write_failure(address=self.unresolved_address),
                 raise
+            except Neo4jError as e:
+                if self.pool and e.invalidates_all_connections():
+                    self.pool.mark_all_stale()
+                raise
         else:
             raise BoltProtocolError("Unexpected response message with signature "
                                     "%02X" % ord(summary_signature), self.unresolved_address)
 
         return len(details), 1
 
-    def _set_defunct(self, error=None):
-        direct_driver = isinstance(self.pool, BoltPool)
-
-        message = ("Failed to read from defunct connection {!r} ({!r})".format(
-            self.unresolved_address, self.server_info.address))
-
-        if error:
-            log.error(str(error))
-        log.error(message)
-        # We were attempting to receive data but the connection
-        # has unexpectedly terminated. So, we need to close the
-        # connection from the client side, and remove the address
-        # from the connection pool.
-        self._defunct = True
-        self.close()
-        if self.pool:
-            self.pool.deactivate(address=self.unresolved_address)
-        # Iterate through the outstanding responses, and if any correspond
-        # to COMMIT requests then raise an error to signal that we are
-        # unable to confirm that the COMMIT completed successfully.
-        for response in self.responses:
-            if isinstance(response, CommitResponse):
-                if error:
-                    raise BoltIncompleteCommitError(message, address=None) from error
-                else:
-                    raise BoltIncompleteCommitError(message, address=None)
-
-        if direct_driver:
-            if error:
-                raise ServiceUnavailable(message) from error
-            else:
-                raise ServiceUnavailable(message)
-        else:
-            if error:
-                raise SessionExpired(message) from error
-            else:
-                raise SessionExpired(message)
-
-    def timedout(self):
-        return 0 <= self._max_connection_lifetime <= perf_counter() - self._creation_timestamp
-
-    def fetch_all(self):
-        """ Fetch all outstanding messages.
-
-        :return: 2-tuple of number of detail messages and number of summary
-                 messages fetched
-        """
-        detail_count = summary_count = 0
-        while self.responses:
-            response = self.responses[0]
-            while not response.complete:
-                detail_delta, summary_delta = self.fetch_message()
-                detail_count += detail_delta
-                summary_count += summary_delta
-        return detail_count, summary_count
-
     def close(self):
         """ Close the connection.
         """
@@ -440,12 +327,12 @@ class Bolt4x0(Bolt):
                 self._append(b"\x02", ())
                 try:
                     self._send_all()
-                except:
+                except (OSError, BoltError, DriverError):
                     pass
             log.debug("[#%04X]  C: <CLOSE>", self.local_port)
             try:
                 self.socket.close()
-            except IOError:
+            except OSError:
                 pass
             finally:
                 self._closed = True
@@ -460,7 +347,7 @@ class Bolt4x0(Bolt):
 class Bolt4x1(Bolt4x0):
     """ Protocol handler for Bolt 4.1.
 
-    This is supported by Neo4j versions 4.1 and 4.2.
+    This is supported by Neo4j versions 4.1 - 4.4.
     """
 
     PROTOCOL_VERSION = Version(4, 1)
@@ -471,16 +358,176 @@ class Bolt4x1(Bolt4x0):
         enables server-side routing to propagate the same behaviour
         through its driver.
         """
-        return {
+        headers = {
             "user_agent": self.user_agent,
-            "routing": self.routing_context,
         }
+        if self.routing_context is not None:
+            headers["routing"] = self.routing_context
+        return headers
 
 
 class Bolt4x2(Bolt4x1):
     """ Protocol handler for Bolt 4.2.
 
-    This is supported by Neo4j version 4.2.
+    This is supported by Neo4j version 4.2 - 4.4.
     """
 
     PROTOCOL_VERSION = Version(4, 2)
+
+
+class Bolt4x3(Bolt4x2):
+    """ Protocol handler for Bolt 4.3.
+
+    This is supported by Neo4j version 4.3 - 4.4.
+    """
+
+    PROTOCOL_VERSION = Version(4, 3)
+
+    def route(self, database=None, imp_user=None, bookmarks=None):
+        if imp_user is not None:
+            raise ConfigurationError(
+                "Impersonation is not supported in Bolt Protocol {!r}. "
+                "Trying to impersonate {!r}.".format(
+                    self.PROTOCOL_VERSION, imp_user
+                )
+            )
+
+        routing_context = self.routing_context or {}
+        log.debug("[#%04X]  C: ROUTE %r %r %r", self.local_port,
+                  routing_context, bookmarks, database)
+        metadata = {}
+        if bookmarks is None:
+            bookmarks = []
+        else:
+            bookmarks = list(bookmarks)
+        self._append(b"\x66", (routing_context, bookmarks, database),
+                     response=Response(self, "route",
+                                       on_success=metadata.update))
+        self.send_all()
+        self.fetch_all()
+        return [metadata.get("rt")]
+
+    def hello(self):
+        def on_success(metadata):
+            self.configuration_hints.update(metadata.pop("hints", {}))
+            self.server_info.update(metadata)
+            if "connection.recv_timeout_seconds" in self.configuration_hints:
+                recv_timeout = self.configuration_hints[
+                    "connection.recv_timeout_seconds"
+                ]
+                if isinstance(recv_timeout, int) and recv_timeout > 0:
+                    self.socket.settimeout(recv_timeout)
+                else:
+                    log.info("[#%04X]  Server supplied an invalid value for "
+                             "connection.recv_timeout_seconds (%r). Make sure "
+                             "the server and network is set up correctly.",
+                             self.local_port, recv_timeout)
+
+        headers = self.get_base_headers()
+        headers.update(self.auth_dict)
+        logged_headers = dict(headers)
+        if "credentials" in logged_headers:
+            logged_headers["credentials"] = "*******"
+        log.debug("[#%04X]  C: HELLO %r", self.local_port, logged_headers)
+        self._append(b"\x01", (headers,),
+                     response=InitResponse(self, "hello",
+                                           on_success=on_success))
+        self.send_all()
+        self.fetch_all()
+        check_supported_server_product(self.server_info.agent)
+
+
+class Bolt4x4(Bolt4x3):
+    """ Protocol handler for Bolt 4.4.
+
+    This is supported by Neo4j version 4.4.
+    """
+
+    PROTOCOL_VERSION = Version(4, 4)
+
+    def route(self, database=None, imp_user=None, bookmarks=None):
+        routing_context = self.routing_context or {}
+        db_context = {}
+        if database is not None:
+            db_context.update(db=database)
+        if imp_user is not None:
+            db_context.update(imp_user=imp_user)
+        log.debug("[#%04X]  C: ROUTE %r %r %r", self.local_port,
+                  routing_context, bookmarks, db_context)
+        metadata = {}
+        if bookmarks is None:
+            bookmarks = []
+        else:
+            bookmarks = list(bookmarks)
+        self._append(b"\x66", (routing_context, bookmarks, db_context),
+                     response=Response(self, "route",
+                                       on_success=metadata.update))
+        self.send_all()
+        self.fetch_all()
+        return [metadata.get("rt")]
+
+    def run(self, query, parameters=None, mode=None, bookmarks=None,
+            metadata=None, timeout=None, db=None, imp_user=None, **handlers):
+        if not parameters:
+            parameters = {}
+        extra = {}
+        if mode in (READ_ACCESS, "r"):
+            # It will default to mode "w" if nothing is specified
+            extra["mode"] = "r"
+        if db:
+            extra["db"] = db
+        if imp_user:
+            extra["imp_user"] = imp_user
+        if bookmarks:
+            try:
+                extra["bookmarks"] = list(bookmarks)
+            except TypeError:
+                raise TypeError("Bookmarks must be provided within an iterable")
+        if metadata:
+            try:
+                extra["tx_metadata"] = dict(metadata)
+            except TypeError:
+                raise TypeError("Metadata must be coercible to a dict")
+        if timeout:
+            try:
+                extra["tx_timeout"] = int(1000 * timeout)
+            except TypeError:
+                raise TypeError("Timeout must be specified as a number of "
+                                "seconds")
+        fields = (query, parameters, extra)
+        log.debug("[#%04X]  C: RUN %s", self.local_port,
+                  " ".join(map(repr, fields)))
+        if query.upper() == u"COMMIT":
+            self._append(b"\x10", fields, CommitResponse(self, "run",
+                                                         **handlers))
+        else:
+            self._append(b"\x10", fields, Response(self, "run", **handlers))
+
+    def begin(self, mode=None, bookmarks=None, metadata=None, timeout=None,
+              db=None, imp_user=None, **handlers):
+        extra = {}
+        if mode in (READ_ACCESS, "r"):
+            # It will default to mode "w" if nothing is specified
+            extra["mode"] = "r"
+        if db:
+            extra["db"] = db
+        if imp_user:
+            extra["imp_user"] = imp_user
+        if bookmarks:
+            try:
+                extra["bookmarks"] = list(bookmarks)
+            except TypeError:
+                raise TypeError("Bookmarks must be provided within an iterable")
+        if metadata:
+            try:
+                extra["tx_metadata"] = dict(metadata)
+            except TypeError:
+                raise TypeError("Metadata must be coercible to a dict")
+        if timeout:
+            try:
+                extra["tx_timeout"] = int(1000 * timeout)
+            except TypeError:
+                raise TypeError("Timeout must be specified as a number of "
+                                "seconds")
+        log.debug("[#%04X]  C: BEGIN %r", self.local_port, extra)
+        self._append(b"\x11", (extra,), Response(self, "begin", **handlers))
