@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- encoding: utf-8 -*-
 
-# Copyright (c) 2002-2020 "Neo4j,"
+# Copyright (c) "Neo4j"
 # Neo4j Sweden AB [http://neo4j.com]
 #
 # This file is part of Neo4j.
@@ -19,29 +19,31 @@
 # limitations under the License.
 
 
-from collections import deque
 from logging import getLogger
 from random import random
-from time import perf_counter, sleep
-from warnings import warn
+from time import (
+    perf_counter,
+    sleep,
+)
 
+from neo4j.api import (
+    READ_ACCESS,
+    WRITE_ACCESS,
+)
 from neo4j.conf import SessionConfig
-from neo4j.api import READ_ACCESS, WRITE_ACCESS
-from neo4j.data import DataHydrator, DataDehydrator
+from neo4j.data import DataHydrator
 from neo4j.exceptions import (
+    ClientError,
+    IncompleteCommit,
     Neo4jError,
     ServiceUnavailable,
-    TransientError,
     SessionExpired,
+    TransientError,
     TransactionError,
-    ClientError,
 )
-from neo4j._exceptions import BoltIncompleteCommitError
 from neo4j.work import Workspace
-from neo4j.work.summary import ResultSummary
-from neo4j.work.transaction import Transaction
 from neo4j.work.result import Result
-from neo4j.io._bolt3 import Bolt3
+from neo4j.work.transaction import Transaction
 
 log = getLogger("neo4j")
 
@@ -59,7 +61,7 @@ class Session(Workspace):
     context. For example::
 
         with driver.session() as session:
-            result = session.run("MATCH (a:Person) RETURN a.name")
+            result = session.run("MATCH (n:Person) RETURN n.name AS name")
             # do something with the result...
 
     :param pool: connection pool instance
@@ -75,10 +77,6 @@ class Session(Workspace):
     # The current auto-transaction result, if any.
     _autoResult = None
 
-    # The set of bookmarks after which the next
-    # :class:`.Transaction` should be carried out.
-    _bookmarks = None
-
     # The state this session is in.
     _state_failed = False
 
@@ -93,7 +91,7 @@ class Session(Workspace):
     def __del__(self):
         try:
             self.close()
-        except OSError:
+        except (OSError, ServiceUnavailable, SessionExpired):
             pass
 
     def __enter__(self):
@@ -104,21 +102,10 @@ class Session(Workspace):
             self._state_failed = True
         self.close()
 
-    def _connect(self, access_mode, database):
+    def _connect(self, access_mode):
         if access_mode is None:
             access_mode = self._config.default_access_mode
-        if self._connection:
-            # TODO: Investigate this
-            # log.warning("FIXME: should always disconnect before connect")
-            self._connection.send_all()
-            self._connection.fetch_all()
-            self._disconnect()
-        self._connection = self._pool.acquire(access_mode=access_mode, timeout=self._config.connection_acquisition_timeout, database=database)
-
-    def _disconnect(self):
-        if self._connection:
-            self._connection.in_use = False
-            self._connection = None
+        super()._connect(access_mode)
 
     def _collect_bookmark(self, bookmark):
         if bookmark:
@@ -130,8 +117,16 @@ class Session(Workspace):
             self._autoResult = None
             self._disconnect()
 
+    def _result_error(self, _):
+        if self._autoResult:
+            self._autoResult = None
+            self._disconnect()
+
     def close(self):
-        """Close the session. This will release any borrowed resources, such as connections, and will roll back any outstanding transactions.
+        """Close the session.
+
+        This will release any borrowed resources, such as connections, and will
+        roll back any outstanding transactions.
         """
         if self._connection:
             if self._autoResult:
@@ -192,7 +187,7 @@ class Session(Workspace):
         :type parameters: dict
         :param kwparameters: additional keyword parameters
         :returns: a new :class:`neo4j.Result` object
-        :type: :class:`neo4j.Result`
+        :rtype: :class:`neo4j.Result`
         """
         if not query:
             raise ValueError("Cannot run an empty query")
@@ -206,15 +201,22 @@ class Session(Workspace):
             self._autoResult._buffer_all()  # This will buffer upp all records for the previous auto-transaction
 
         if not self._connection:
-            self._connect(self._config.default_access_mode, database=self._config.database)
+            self._connect(self._config.default_access_mode)
         cx = self._connection
         protocol_version = cx.PROTOCOL_VERSION
         server_info = cx.server_info
 
         hydrant = DataHydrator()
 
-        self._autoResult = Result(cx, hydrant, self._config.fetch_size, self._result_closed)
-        self._autoResult._run(query, parameters, self._config.database, self._config.default_access_mode, self._bookmarks, **kwparameters)
+        self._autoResult = Result(
+            cx, hydrant, self._config.fetch_size, self._result_closed,
+            self._result_error
+        )
+        self._autoResult._run(
+            query, parameters, self._config.database,
+            self._config.impersonated_user, self._config.default_access_mode,
+            self._bookmarks, **kwparameters
+        )
 
         return self._autoResult
 
@@ -243,10 +245,23 @@ class Session(Workspace):
             self._transaction = None
             self._disconnect()
 
-    def _open_transaction(self, *, access_mode, database, metadata=None, timeout=None):
-        self._connect(access_mode=access_mode, database=database)
-        self._transaction = Transaction(self._connection, self._config.fetch_size, self._transaction_closed_handler)
-        self._transaction._begin(database, self._bookmarks, access_mode, metadata, timeout)
+    def _transaction_error_handler(self, _):
+        if self._transaction:
+            self._transaction = None
+            self._disconnect()
+
+    def _open_transaction(self, *, access_mode, metadata=None,
+                          timeout=None):
+        self._connect(access_mode=access_mode)
+        self._transaction = Transaction(
+            self._connection, self._config.fetch_size,
+            self._transaction_closed_handler,
+            self._transaction_error_handler
+        )
+        self._transaction._begin(
+            self._config.database, self._config.impersonated_user,
+            self._bookmarks, access_mode, metadata, timeout
+        )
 
     def begin_transaction(self, metadata=None, timeout=None):
         """ Begin a new unmanaged transaction. Creates a new :class:`.Transaction` within this session.
@@ -263,7 +278,7 @@ class Session(Workspace):
         :type metadata: dict
 
         :param timeout:
-            the transaction timeout in milliseconds.
+            the transaction timeout in seconds.
             Transactions that execute longer than the configured timeout will be terminated by the database.
             This functionality allows to limit query/transaction execution time.
             Specified timeout overrides the default timeout configured in the database using ``dbms.transaction.timeout`` setting.
@@ -283,7 +298,8 @@ class Session(Workspace):
         if self._transaction:
             raise TransactionError("Explicit transaction already open")
 
-        self._open_transaction(access_mode=self._config.default_access_mode, database=self._config.database, metadata=metadata, timeout=timeout)
+        self._open_transaction(access_mode=self._config.default_access_mode,
+                               metadata=metadata, timeout=timeout)
 
         return self._transaction
 
@@ -303,15 +319,17 @@ class Session(Workspace):
 
         while True:
             try:
-                self._open_transaction(access_mode=access_mode, database=self._config.database, metadata=metadata, timeout=timeout)
+                self._open_transaction(access_mode=access_mode, metadata=metadata, timeout=timeout)
                 tx = self._transaction
                 try:
                     result = transaction_function(tx, *args, **kwargs)
                 except Exception:
-                    tx.rollback()
+                    tx.close()
                     raise
                 else:
                     tx.commit()
+            except IncompleteCommit:
+                raise
             except (ServiceUnavailable, SessionExpired) as error:
                 errors.append(error)
                 self._disconnect()
@@ -344,14 +362,34 @@ class Session(Workspace):
 
         Example::
 
-            def do_cypher(tx, cypher):
+            def do_cypher_tx(tx, cypher):
                 result = tx.run(cypher)
-                result.consume()
-                return 1
+                values = []
+                for record in result:
+                    values.append(record.values())
+                return values
 
-            value = session.read_transaction(do_cypher, "RETURN 1")
+            with driver.session() as session:
+                values = session.read_transaction(do_cypher_tx, "RETURN 1 AS x")
 
-        :param transaction_function: a function that takes a transaction as an argument and does work with the transaction. `tx_function(tx, \*args, \*\*kwargs)`
+        Example::
+
+            def get_two_tx(tx):
+                result = tx.run("UNWIND [1,2,3,4] AS x RETURN x")
+                values = []
+                for ix, record in enumerate(result):
+                    if x > 1:
+                        break
+                    values.append(record.values())
+                # discard the remaining records if there are any
+                summary = result.consume()
+                # use the summary for logging etc.
+                return values
+
+            with driver.session() as session:
+                values = session.read_transaction(get_two_tx)
+
+        :param transaction_function: a function that takes a transaction as an argument and does work with the transaction. `tx_function(tx, *args, **kwargs)`
         :param args: arguments for the `transaction_function`
         :param kwargs: key word arguments for the `transaction_function`
         :return: a result as returned by the given unit of work
@@ -367,14 +405,16 @@ class Session(Workspace):
 
         Example::
 
-            def do_cypher(tx, cypher):
-                result = tx.run(cypher)
-                result.consume()
-                return 1
+            def create_node_tx(tx, name):
+                result = tx.run("CREATE (n:NodeExample { name: $name }) RETURN id(n) AS node_id", name=name)
+                record = result.single()
+                return record["node_id"]
 
-            value = session.write_transaction(do_cypher, "RETURN 1")
+            with driver.session() as session:
+                node_id = session.write_transaction(create_node_tx, "example")
 
-        :param transaction_function: a function that takes a transaction as an argument and does work with the transaction. `tx_function(tx, \*args, \*\*kwargs)`
+
+        :param transaction_function: a function that takes a transaction as an argument and does work with the transaction. `tx_function(tx, *args, **kwargs)`
         :param args: key word arguments for the `transaction_function`
         :param kwargs: key word arguments for the `transaction_function`
         :return: a result as returned by the given unit of work
@@ -390,7 +430,7 @@ class Query:
     :param metadata: metadata attached to the query.
     :type metadata: dict
     :param timeout: seconds.
-    :type timeout: int
+    :type timeout: float or :const:`None`
     """
     def __init__(self, text, metadata=None, timeout=None):
         self.text = text
@@ -407,9 +447,13 @@ def unit_of_work(metadata=None, timeout=None):
 
     For example, a timeout may be applied::
 
+        from neo4j import unit_of_work
+
         @unit_of_work(timeout=100)
-        def count_people(tx):
-            return tx.run("MATCH (a:Person) RETURN count(a)").single().value()
+        def count_people_tx(tx):
+            result = tx.run("MATCH (a:Person) RETURN count(a) AS persons")
+            record = result.single()
+            return record["persons"]
 
     :param metadata:
         a dictionary with metadata.
@@ -419,12 +463,12 @@ def unit_of_work(metadata=None, timeout=None):
     :type metadata: dict
 
     :param timeout:
-        the transaction timeout in milliseconds.
+        the transaction timeout in seconds.
         Transactions that execute longer than the configured timeout will be terminated by the database.
         This functionality allows to limit query/transaction execution time.
         Specified timeout overrides the default timeout configured in the database using ``dbms.transaction.timeout`` setting.
         Value should not represent a duration of zero or negative duration.
-    :type timeout: int
+    :type timeout: float or :const:`None`
     """
 
     def wrapper(f):
